@@ -7,11 +7,19 @@ import os
 import inspect
 import warnings
 import ctypes
+from contextlib import nullcontext
+from collections import defaultdict
 from . import irtools
+from .errors import UnsupportedError
 from .typesystem import Type, get_signature
 from .thrift import Server, Dispatcher, dispatchermethod, Data, Client
-from .utils import get_local_ip
+from .utils import get_local_ip, UNSPECIFIED
 from .targetinfo import TargetInfo
+from .rbclib import tracing_allocator
+# XXX WIP: the OmnisciCompilerPipeline is no longer omnisci-specific because
+# we support Arrays even without omnisci, so it must be renamed and moved
+# somewhere elsef
+from .omnisci_backend import OmnisciCompilerPipeline
 
 
 def isfunctionlike(obj):
@@ -60,7 +68,7 @@ def extract_templates(options):
     return new_options, templates
 
 
-class Signature(object):
+class Signature:
     """Signature decorator for Python functions.
 
     A Signature decorator may contain many signature objects
@@ -110,9 +118,20 @@ class Signature(object):
         assert not self.signature_templates
         return sig
 
+    def __repr__(self):
+        return f'{type(self).__name__}({str(self)})'
+
     def __str__(self):
-        lst = ["'%s'" % (s,) for s in self.signatures]
-        return '%s(%s)' % (self.__class__.__name__, ', '.join(lst))
+        lst = []
+        for t in self.signatures:
+            s = str(t)
+            for k, types in self.signature_templates.get(t, {}).items():
+                s += f', {k}={"|".join(map(str, types))}'
+            devices = self.signature_devices.get(t, [])
+            if devices:
+                s += f', device={"|".join(devices)}'
+            lst.append(repr(s))
+        return f'{"; ".join(lst)}'
 
     def __call__(self, obj, **options):
         """Decorate signatures or a function.
@@ -204,19 +223,19 @@ class Signature(object):
         ftype = None
         match_penalty = None
         available_types = self.normalized(func).signatures
+
         for typ in available_types:
-            penalty = typ.match(atypes)
+            sig = self.remotejit.caller_signature(typ)
+            penalty = sig.match(atypes)
             if penalty is not None:
                 if ftype is None or penalty < match_penalty:
                     ftype = typ
                     match_penalty = penalty
-        if ftype is None:
-            satypes = ', '.join(map(str, atypes))
-            available = '; '.join(map(str, available_types))
-            raise TypeError(
-                f'found no matching function type to given argument types'
-                f' `{satypes}`. Available function types: {available}')
-        return ftype
+        return ftype, match_penalty
+
+    def add(self, sig):
+        if sig not in self.signatures:
+            self.signatures.append(sig)
 
     def normalized(self, func=None):
         """Return a copy of Signature object where all signatures are
@@ -263,18 +282,18 @@ class Signature(object):
             if not sig.is_concrete:
                 for csig in sig.apply_templates(templates):
                     assert isinstance(csig, Type), (sig, csig, type(csig))
-                    if csig not in signature.signatures:
-                        signature.signatures.append(csig)
+                    csig = self.remotejit.normalize_function_type(csig)
+                    signature.add(csig)
             else:
-                if sig not in signature.signatures:
-                    signature.signatures.append(sig)
+                sig = self.remotejit.normalize_function_type(sig)
+                signature.add(sig)
         if fsig is not None and fsig.is_complete:
-            if fsig not in signature.signatures:
-                signature.signatures.append(fsig)
+            fsig = self.remotejit.normalize_function_type(fsig)
+            signature.add(fsig)
         return signature
 
 
-class Caller(object):
+class Caller:
     """Remote JIT caller, holds the decorated function that can be
     executed remotely.
     """
@@ -310,14 +329,20 @@ class Caller(object):
         """Return Caller instance that executes function calls on the local
         host. Useful for debugging.
         """
+        if not self.remotejit.supports_local_caller:
+            msg = (
+                "Cannot create a local `Caller` when using "
+                f"{type(self.remotejit).__name__}."
+            )
+            raise UnsupportedError(msg)
+
         return Caller(self.func, self.signature.local)
 
     def __repr__(self):
-        return '%s(%s, %s, local=%s)' % (type(self).__name__, self.func,
-                                         self.signature, self.local)
+        return f"{type(self).__name__}({self.func}, {self.signature!r})"
 
     def __str__(self):
-        return self.describe()
+        return f"{self.func.__name__}[{self.signature}]"
 
     def describe(self):
         """Return LLVM IRs of all target devices.
@@ -336,6 +361,7 @@ class Caller(object):
                     llvm_module, succesful_fids = irtools.compile_to_LLVM(
                         [(self.func, signatures_map)],
                         target_info,
+                        pipeline_class=OmnisciCompilerPipeline,
                         debug=self.remotejit.debug)
                     lst.append(str(llvm_module))
         lst.append(f'{"":-^80}')
@@ -348,29 +374,132 @@ class Caller(object):
 
     # RBC user-interface
 
-    def __call__(self, *arguments, **options):
+    def __call__(self, *arguments, device=UNSPECIFIED, hold=UNSPECIFIED):
         """Return the result of a remote JIT compiled function call.
         """
-        device = options.get('device')
-        targets = self.remotejit.targets
-        if device is None:
-            if len(targets) > 1:
-                raise TypeError(
-                    f'specifying device is required when target has more than'
-                    f' one device. Available devices: {", ".join(targets)}')
-            device = tuple(targets)[0]
-        target_info = targets[device]
-        with target_info:
-            atypes = tuple(map(Type.fromvalue, arguments))
-            ftype = self.signature.best_match(self.func, atypes)
-            key = self.func.__name__, ftype
-            if key not in self._is_compiled:
-                self.remotejit.remote_compile(self.func, ftype, target_info)
-                self._is_compiled.add(key)
-            return self.remotejit.remote_call(self.func, ftype, arguments)
+        caller = self.remotejit.get_caller(self.func.__name__)
+        return caller(*arguments, device=device, hold=hold)
 
 
-class RemoteJIT(object):
+class RemoteDispatcher:
+    """A collection of Caller instances holding functions with a common name.
+    """
+    def __init__(self, name, callers):
+        self.name = name
+        assert callers  # at least one caller must be specified
+        self.remotejit = callers[0].remotejit
+        self.callers = callers
+
+    def __repr__(self):
+        lst = [str(caller.signature) for caller in self.callers]
+        return f'{type(self).__name__}({self.name!r}, [{", ".join(lst)}])'
+
+    def __str__(self):
+        lst = [str(caller.signature) for caller in self.callers]
+        return f'{self.name}[{", ".join(lst)}]'
+
+    def __call__(self, *arguments, device=UNSPECIFIED, hold=UNSPECIFIED):
+        """Perform remote call with given arguments.
+
+        If `hold` is True, return an object that encapsulates the
+        remote call to postpone the remote execution.
+        """
+        if hold is UNSPECIFIED:
+            hold = self.remotejit.default_remote_call_hold
+
+        penalty_device_caller_ftype = []
+        atypes = None
+        for device_, target_info in self.remotejit.targets.items():
+            if device is not UNSPECIFIED and device != device_:
+                continue
+            with target_info:
+                atypes = self.remotejit.get_types(*arguments)
+                for caller_id, caller in enumerate(self.callers):
+                    with Type.alias(**self.remotejit.typesystem_aliases):
+                        ftype, penalty = caller.signature.best_match(caller.func, atypes)
+                        if ftype is None:
+                            continue
+                        penalty_device_caller_ftype.append((penalty, device_, caller_id, ftype))
+        if atypes is None:
+            raise ValueError(f'no target info found for given device {device}')
+
+        penalty_device_caller_ftype.sort()
+
+        if not penalty_device_caller_ftype:
+            available_types_devices = defaultdict(set)
+            for device_, target_info in self.remotejit.targets.items():
+                if device is not UNSPECIFIED and device != device_:
+                    continue
+                with target_info:
+                    for caller in self.callers:
+                        with Type.alias(**self.remotejit.typesystem_aliases):
+                            for t in caller.signature.normalized(caller.func).signatures:
+                                available_types_devices[t].add(device_)
+            lines = self.remotejit._format_available_function_types(available_types_devices)
+            available = '\n    ' + '\n    '.join(lines)
+            satypes = ', '.join(map(str, atypes))
+            raise TypeError(
+                f'found no matching function signature to given argument types:'
+                f'\n    ({satypes}) -> ...\n  available function signatures:{available}')
+
+        _, device, caller_id, ftype = penalty_device_caller_ftype[0]
+        target_info = self.remotejit.targets[device]
+        caller = self.callers[caller_id]
+        r = self.remotejit.remote_call_capsule_cls(caller, target_info, ftype, arguments)
+
+        return r if hold else r.execute()
+
+
+class RemoteCallCapsule:
+    """Encapsulates remote call execution.
+    """
+
+    use_execute_cache = False
+
+    def __init__(self, caller, target_info, ftype, arguments):
+        self.caller = caller
+        self.target_info = target_info
+        self.ftype = ftype
+        self.arguments = arguments
+        self._execute_cache = UNSPECIFIED
+
+    @property
+    def __typesystem_type__(self):
+        """The typesystem Type instance of the return value of the remote
+        call.
+        """
+        return self.caller.remotejit.caller_signature(self.ftype)[0]
+
+    def __repr__(self):
+        return (f'{type(self).__name__}({self.caller!r}, {self.target_info},'
+                f' {self.ftype}, {self.arguments})')
+
+    def __str__(self):
+        return f'{self.execute(hold=True)}'
+
+    def execute(self, hold=False):
+        """Trigger the remote call execution.
+
+        When `hold` is True, return an object that represents the
+        remote call.
+        """
+        if not hold:
+            if self.use_execute_cache and self._execute_cache is not UNSPECIFIED:
+                return self._execute_cache
+
+            key = self.caller.func.__name__, self.ftype
+            if key not in self.caller._is_compiled:
+                self.caller.remotejit.remote_compile(
+                    self.caller.func, self.ftype, self.target_info)
+                self.caller._is_compiled.add(key)
+        result = self.caller.remotejit.remote_call(self.caller.func, self.ftype,
+                                                   self.arguments, hold=hold)
+        if not hold and self.use_execute_cache:
+            self._execute_cache = result
+        return result
+
+
+class RemoteJIT:
     """RemoteJIT is a decorator generator for user functions to be
     remotely JIT compiled.
 
@@ -401,8 +530,18 @@ class RemoteJIT(object):
 
     thrift_content = None
 
+    typesystem_aliases = dict()
+
+    remote_call_capsule_cls = RemoteCallCapsule
+
+    # Some callers cannot be called locally
+    supports_local_caller = True
+
+    # Should calling RemoteDispatcher hold the execution:
+    default_remote_call_hold = False
+
     def __init__(self, host='localhost', port=11532,
-                 local=False, debug=False):
+                 local=False, debug=False, use_tracing_allocator=False):
         """Construct remote JIT function decorator.
 
         The decorator is re-usable for different functions.
@@ -417,11 +556,17 @@ class RemoteJIT(object):
           When True, use local client. Useful for debugging.
         debug : bool
           When True, output debug messages.
+        use_tracing_allocator : bool
+          When True, enable the automatic detection of memory leaks.
         """
         if host == 'localhost':
             host = get_local_ip()
 
+        if use_tracing_allocator and not local:
+            raise ValueError('use_tracing_allocator=True can be used only with local=True')
+
         self.debug = debug
+        self.use_tracing_allocator = use_tracing_allocator
         self.host = host
         self.port = int(port)
         self.server_process = None
@@ -433,7 +578,8 @@ class RemoteJIT(object):
         self._targets = None
 
         if local:
-            self._client = LocalClient(debug=debug)
+            self._client = LocalClient(debug=debug,
+                                       use_tracing_allocator=use_tracing_allocator)
         else:
             self._client = None
 
@@ -444,11 +590,26 @@ class RemoteJIT(object):
         return localjit
 
     def add_caller(self, caller):
-        self._callers.append(caller)
+        name = caller.func.__name__
+        for c in self._callers:
+            if c.name == name:
+                c.callers.append(caller)
+                break
+        else:
+            self._callers.append(RemoteDispatcher(name, [caller]))
         self.discard_last_compile()
 
     def get_callers(self):
-        return self._callers
+        callers = []
+        for c in self._callers:
+            callers.extend(c.callers)
+        return callers
+
+    def get_caller(self, name):
+        for c in self._callers:
+            if c.name == name:
+                return c
+        return
 
     def reset(self):
         """Drop all callers definitions and compilation results.
@@ -548,12 +709,12 @@ class RemoteJIT(object):
             self._targets = self.retrieve_targets()
         return self._targets
 
-    def __call__(self, *signatures, **options):
+    def __call__(self, *signatures, devices=None, local=False, **templates):
         """Define a remote JIT function signatures and template.
 
         Parameters
         ----------
-        signatures : tuple
+        signatures : str or object
           Specify signatures of a remote JIT function, or a Python
           function as a template from which the remote JIT function
           will be compiled.
@@ -562,8 +723,9 @@ class RemoteJIT(object):
         ------------------
         local : bool
         devices : list
-          Specify device names for the given set of signatures.
-        templates : dict
+          Specify device names for the given set of signatures. Possible
+          values are 'cpu', 'gpu'.
+        templates : dict(str, list(str))
           Specify template types mapping.
 
         Returns
@@ -580,12 +742,20 @@ class RemoteJIT(object):
         or any other object that can be converted to function type,
         see `Type.fromobject` for more information.
         """
-        if options.get('local'):
+        if local:
             s = Signature(self.local)
         else:
             s = Signature(self)
-        devices = options.get('devices')
-        options, templates = extract_templates(options)
+        options = dict(
+            local=local,
+            devices=devices,
+            templates=templates.get('templates') or templates
+        )
+        if devices is not None and not {'cpu', 'gpu'}.issuperset(devices):
+            raise ValueError("'devices' can only be a list with possible "
+                             f"values 'cpu', 'gpu' but got {devices}")
+
+        _, templates = extract_templates(options)
         for sig in signatures:
             s = s(sig, devices=devices, templates=templates)
         return s
@@ -646,8 +816,12 @@ class RemoteJIT(object):
         """
         if self.debug:
             print(f'remote_compile({func}, {ftype})')
-        llvm_module, succesful_fids = irtools.compile_to_LLVM(
-            [(func, {0: ftype})], target_info, debug=self.debug)
+        with target_info:
+            llvm_module, succesful_fids = irtools.compile_to_LLVM(
+                [(func, {0: ftype})],
+                target_info,
+                pipeline_class=OmnisciCompilerPipeline,
+                debug=self.debug)
         ir = str(llvm_module)
         mangled_signatures = ';'.join([s.mangle() for s in [ftype]])
         response = self.client(remotejit=dict(
@@ -655,18 +829,25 @@ class RemoteJIT(object):
         assert response['remotejit']['compile'], response
         return llvm_module
 
-    def remote_call(self, func, ftype: Type, arguments: tuple):
+    def remote_call(self, func, ftype: Type, arguments: tuple, hold=False):
         """Call function remotely on given arguments.
 
         The input function `func` is called remotely by sending the
         arguments data to remote host where the previously compiled
         function (see `remote_compile` method) is applied to the
         arguments, and the result is returned to local process.
+
+        If `hold` is True then return an object that specifies remote
+        call but does not execute it. The type of return object is
+        custom to particular RemoteJIT specialization.
         """
         if self.debug:
             print(f'remote_call({func}, {ftype}, {arguments})')
         fullname = func.__name__ + ftype.mangle()
-        response = self.client(remotejit=dict(call=(fullname, arguments)))
+        call = dict(call=(fullname, arguments))
+        if hold:
+            return call
+        response = self.client(remotejit=call)
         return response['remotejit']['call']
 
     def python(self, statement):
@@ -675,6 +856,22 @@ class RemoteJIT(object):
         response = self.client(remotejit=dict(python=(statement,)))
         return response['remotejit']['python']
 
+    def normalize_function_type(self, ftype: Type):
+        """Apply RemoteJIT specific hooks to normalized function Type.
+
+        Parameters
+        ----------
+        ftype: Type
+          typesystem type of a function
+
+        Returns
+        -------
+        ftype: Type
+          typesystem type of a function with normalization hools applied
+        """
+        assert ftype.is_function, ftype
+        return ftype
+
     def preprocess_callable(self, func):
         """Preprocess func to be used as a remotejit function definition.
 
@@ -689,13 +886,54 @@ class RemoteJIT(object):
         """
         return func
 
+    def caller_signature(self, signature: Type):
+        """Return signature of a caller.
+
+        Parameters
+        ----------
+        signature: Type
+          Signature of function implementation
+
+        Returns
+        -------
+        signature: Type
+          Signature of function caller
+        """
+        return signature
+
+    def get_types(self, *values):
+        """Convert values to the corresponding typesystem types.
+        """
+        return tuple(map(Type.fromvalue, values))
+
+    def format_type(self, typ: Type):
+        """Convert typesystem type to formatted string.
+        """
+        return str(typ)
+
+    def _format_available_function_types(self, available_types_devices):
+        all_devices = set()
+        list(map(all_devices.update, available_types_devices.values()))
+        lines = []
+        for typ, devices in available_types_devices.items():
+            sig = self.caller_signature(typ)
+            d = '  ' + '|'.join(devices) + ' only' if len(devices) != len(all_devices) else ''
+            s = self.format_type(sig)
+            t = self.format_type(typ)
+            if sig == typ:
+                lines.append(f'{s}{d}')
+            else:
+                lines.append(f'{s}{d}\n    - {t}')
+        return lines
+
 
 class DispatcherRJIT(Dispatcher):
     """Implements remotejit service methods.
     """
 
-    def __init__(self, server, debug=False):
+    def __init__(self, server, debug=False, use_tracing_allocator=False):
         super().__init__(server, debug=debug)
+        self.use_tracing_allocator = use_tracing_allocator
         self.compiled_functions = dict()
         self.engines = dict()
         self.python_globals = dict()
@@ -710,7 +948,11 @@ class DispatcherRJIT(Dispatcher):
         info : dict
           Map of target devices and their properties.
         """
-        target_info = TargetInfo.host()
+        if self.use_tracing_allocator:
+            target_info = TargetInfo.host(name='host_cpu_tracing_allocator',
+                                          use_tracing_allocator=True)
+        else:
+            target_info = TargetInfo.host()
         target_info.set('has_numba', True)
         target_info.set('has_cpython', True)
         return dict(cpu=target_info.tojson())
@@ -760,6 +1002,16 @@ class DispatcherRJIT(Dispatcher):
         arguments : tuple
           Specify the arguments to the function.
         """
+        # if we are using a tracing allocator, automatically detect memory leaks
+        # at each call.
+        if self.use_tracing_allocator:
+            leak_detector = tracing_allocator.new_leak_detector()
+        else:
+            leak_detector = nullcontext()
+        with leak_detector:
+            return self._do_call(fullname, arguments)
+
+    def _do_call(self, fullname, arguments):
         if self.debug:
             print(f'call({fullname}, {arguments})')
         ef = self.compiled_functions.get(fullname)
@@ -818,14 +1070,15 @@ class DebugDispatcherRJIT(DispatcherRJIT):
     debug = True
 
 
-class LocalClient(object):
+class LocalClient:
     """Pretender of thrift.Client.
 
     All calls will be made in a local process. Useful for debbuging.
     """
 
-    def __init__(self, debug=False):
-        self.dispatcher = DispatcherRJIT(None, debug=debug)
+    def __init__(self, debug=False, use_tracing_allocator=False):
+        self.dispatcher = DispatcherRJIT(None, debug=debug,
+                                         use_tracing_allocator=use_tracing_allocator)
 
     def __call__(self, **services):
         results = {}
