@@ -1,6 +1,8 @@
 import os
 from collections import defaultdict
 from rbc.omnisci_backend import Array
+from rbc.errors import OmnisciServerError
+from rbc.stdlib import array_api
 from numba import types as nb_types
 import pytest
 
@@ -48,64 +50,6 @@ def omnisci():
         print('%s in deardown' % (type(msg)))
 
 
-def _test_get_array_size_ir1(omnisci):
-    omnisci.reset()
-    # register an empty set of UDFs in order to avoid unregistering
-    # UDFs created directly from LLVM IR strings when executing SQL
-    # queries:
-    omnisci.register()
-
-    device_params = omnisci.thrift_call('get_device_parameters')
-    cpu_target_triple = device_params['cpu_triple']
-    cpu_target_datalayout = "e-m:e-i64:64-f80:128-n8:16:32:64-S128"
-
-    # The following are codelets from clang --emit-llvm output with
-    # attributes removed and change of function names:
-    Array_getSize_i32_ir = '''\
-define dso_local i64 @Array_getSize_i32(%struct.Array*) align 2 {
-  %2 = alloca %struct.Array*, align 8
-  store %struct.Array* %0, %struct.Array** %2, align 8
-  %3 = load %struct.Array*, %struct.Array** %2, align 8
-  %4 = getelementptr inbounds %struct.Array, %struct.Array* %3, i32 0, i32 1
-  %5 = load i64, i64* %4, align 8
-  ret i64 %5
-}
-'''
-
-    array_sz_i32_ir = '''\
-define dso_local i32 @array_sz_int32(%struct.Array* byval align 8) {
-  %2 = call i64 @Array_getSize_i32(%struct.Array* %0)
-  %3 = trunc i64 %2 to i32
-  ret i32 %3
-}'''
-
-    ast_signatures = "array_sz_int32 'int32_t(Array<int32_t>)'"
-
-    device_ir_map = dict()
-    device_ir_map['cpu'] = f'''\
-target datalayout = "{cpu_target_datalayout}"
-target triple = "{cpu_target_triple}"
-
-%struct.Array = type {{ i32*, i64, i8 }}
-
-{Array_getSize_i32_ir}
-
-{array_sz_i32_ir}
-'''
-
-    omnisci.thrift_call('register_runtime_udf',
-                        omnisci.session_id,
-                        ast_signatures, device_ir_map)
-
-    desrc, result = omnisci.sql_execute(
-        f'select i4, array_sz_int32(i4) from {omnisci.table_name}')
-    for a, sz in result:
-        assert len(a) == sz
-
-
-@pytest.mark.skipif(available_version[:2] < (5, 2),
-                    reason="test requires 5.2 or newer (got %s)" % (
-                        available_version,))
 @pytest.mark.parametrize('c_name', ['int8_t i1', 'int16_t i2', 'int32_t i4', 'int64_t i8',
                                     'float f4', 'double f8'])
 @pytest.mark.parametrize('device', ['cpu', 'gpu'])
@@ -310,10 +254,6 @@ def test_even_sum(omnisci):
 
 
 def test_array_setitem(omnisci):
-    if omnisci.has_cuda and omnisci.version < (5, 5):
-        pytest.skip(
-            'test_array_setitem: crashes CUDA enabled omniscidb server'
-            ' [rbc issue 72]')
     omnisci.reset()
 
     @omnisci('double(double[], int32)')
@@ -334,20 +274,15 @@ def test_array_setitem(omnisci):
 
 
 def test_array_constructor_noreturn(omnisci):
-    if omnisci.has_cuda and omnisci.version < (5, 5):
-        pytest.skip(
-            'crashes CUDA enabled omniscidb server [issue 94]')
-
     omnisci.reset()
 
     from rbc.omnisci_backend import Array
-    from numba import types
 
     @omnisci('float64(int32)')
     def array_noreturn(size):
-        a = Array(size, types.float64)
-        b = Array(size, types.float64)
-        c = Array(size, types.float64)
+        a = Array(size, nb_types.float64)
+        b = Array(size, nb_types.float64)
+        c = Array(size, nb_types.float64)
         for i in range(size):
             a[i] = b[i] = c[i] = i + 3.0
         s = 0.0
@@ -362,21 +297,16 @@ def test_array_constructor_noreturn(omnisci):
 
 
 def test_array_constructor_return(omnisci):
-    if available_version[:3] == (5, 3, 1):
-        pytest.skip(
-            'crashes CPU-only omniscidb server v 5.3.1 [issue 112]')
-
     omnisci.reset()
 
     from rbc.omnisci_backend import Array
-    from numba import types
     from rbc.externals.stdio import printf
 
     @omnisci('float64[](int32)')
     def array_return(size):
         printf("entering array_return(%i)\n", size)
-        a = Array(size, types.float64)
-        b = Array(size, types.float64)
+        a = Array(size, nb_types.float64)
+        b = Array(size, nb_types.float64)
         for i in range(size):
             a[i] = float(i)
             b[i] = float(size - i - 1)
@@ -399,11 +329,10 @@ def test_array_constructor_len(omnisci):
     omnisci.reset()
 
     from rbc.omnisci_backend import Array
-    from numba import types
 
     @omnisci('int64(int32)')
     def array_len(size):
-        a = Array(size, types.float64)
+        a = Array(size, nb_types.float64)
         return len(a)
 
     query = 'select array_len(30)'
@@ -413,10 +342,6 @@ def test_array_constructor_len(omnisci):
 
 
 def test_array_constructor_getitem(omnisci):
-    if omnisci.has_cuda and omnisci.version < (5, 5):
-        pytest.skip(
-            'test_array_constructor_getitem: crashes CUDA enabled omniscidb'
-            ' server [rbc issue 72]')
     omnisci.reset()
 
     from rbc.omnisci_backend import Array
@@ -461,18 +386,15 @@ inps = [('int32', 'i4', 'trunc'), ('int32', 'i4', 'sext'),
 def test_issue197(omnisci, typ, col, suffix):
     omnisci.reset()
 
-    import rbc.omnisci_backend as np
-    from numba import types
-
     cast = dict(
-        trunc=types.int64,
-        sext=types.int8,
-        zext=types.uint8,
-        fptrunc=types.float64,
-        fpext=types.float32)[suffix]
+        trunc=nb_types.int64,
+        sext=nb_types.int8,
+        zext=nb_types.uint8,
+        fptrunc=nb_types.float64,
+        fpext=nb_types.float32)[suffix]
 
     def fn_issue197(x):
-        y = np.zeros_like(x)
+        y = array_api.zeros_like(x)
         for i in range(len(x)):
             y[i] = cast(x[i] + 3)
         return y
@@ -494,11 +416,9 @@ def test_issue197(omnisci, typ, col, suffix):
 def test_issue197_bool(omnisci):
     omnisci.reset()
 
-    import rbc.omnisci_backend as np
-
     @omnisci('bool[](bool[])')
     def fn_issue197_bool(x):
-        y = np.zeros_like(x)
+        y = array_api.zeros_like(x)
         for i in range(len(x)):
             y[i] = bool(x[i])
         return y
@@ -526,3 +446,51 @@ def test_issue109(omnisci):
 
     _, result = omnisci.sql_execute('select issue109(3);')
     assert list(result) == [([0.0, 1.0, 2.0, 3.0, 4.0],)]
+
+
+def test_issue77(omnisci):
+
+    @omnisci('int64[]()')
+    def issue77():
+        a = Array(5, 'int64')
+        a.fill(1)
+        return a
+
+    if omnisci.version[:2] >= (5, 8):
+        _, result = omnisci.sql_execute('select issue77();')
+        assert list(result)[0][0] == [1, 1, 1, 1, 1]
+    else:
+        with pytest.raises(OmnisciServerError) as exc:
+            _, result = omnisci.sql_execute('select issue77();')
+
+        assert exc.match('Could not bind issue77()')
+
+
+def test_array_dtype(omnisci):
+    table = omnisci.table_name
+
+    @omnisci('T(T[])', T=['int32', 'int64'])
+    def array_dtype_fn(x):
+        if x.dtype == nb_types.int32:
+            return 32
+        else:
+            return 64
+
+    for col, r in (('i4', 32), ('i8', 64)):
+        _, result = omnisci.sql_execute(f'select array_dtype_fn({col}) from {table}')
+        assert list(result) == [(r,)] * 5
+
+
+def test_array_enumerate(omnisci):
+    table = omnisci.table_name
+
+    @omnisci('T(T[])', T=['int32'])
+    def array_enumerate(x):
+        s = 0
+        for i, e in enumerate(x):
+            s += e
+        return s
+
+    _, result = omnisci.sql_execute(f'select i4, array_enumerate(i4) from {table}')
+    for arr, s in result:
+        assert sum(arr) == s
