@@ -9,6 +9,9 @@ available_version, reason = rbc_omnisci.is_available()
 pytestmark = pytest.mark.skipif(not available_version, reason=reason)
 
 
+NUMERIC_TYPES = ['int8', 'int16', 'int32', 'int64', 'float32', 'float64']
+
+
 @pytest.fixture(scope='module')
 def omnisci():
     for o in omnisci_fixture(globals()):
@@ -19,18 +22,29 @@ def omnisci():
 ndarray_methods = [
     ('fill', (5, 4), [4.0, 4.0, 4.0, 4.0, 4.0]),
     ('max', (5, 4.0), 4.0),
-    ('max_empty', (0, ), -128),
-    ('max_initial', (5, 4.0, 30.0), 30.0),
+    ('max_empty_int8', (0, ), np.iinfo(np.int8).min + 1),
+    ('max_empty_int16', (0, ), np.iinfo(np.int16).min + 1),
+    ('max_empty_int32', (0, ), np.iinfo(np.int32).min + 1),
+    ('max_empty_int64', (0, ), np.iinfo(np.int64).min + 1),
+    ('max_empty_float32', (0, ), np.finfo(np.float32).min),
+    ('max_empty_float64', (0, ), np.finfo(np.float64).min),
+    ('mean', (5, 2), 2.0),
     ('mean', (5, 2.0), 2.0),
-    ('mean_empty_float', (0, ), np.nan),
-    ('mean_empty_int', (0, ), 0),
+    ('mean_empty_int8', (0, ), np.nan),
+    ('mean_empty_int16', (0, ), np.nan),
+    ('mean_empty_int32', (0, ), np.nan),
+    ('mean_empty_int64', (0, ), np.nan),
+    ('mean_empty_float32', (0, ), np.nan),
+    ('mean_empty_float64', (0, ), np.nan),
     ('min', (5, 4.0), 4.0),
-    ('min_empty', (0, ), 32767),
-    ('min_initial', (5, 4.0, -3.0), -3.0),
+    ('min_empty_int8', (0, ), np.iinfo(np.int8).max),
+    ('min_empty_int16', (0, ), np.iinfo(np.int16).max),
+    ('min_empty_int32', (0, ), np.iinfo(np.int32).max),
+    ('min_empty_int64', (0, ), np.iinfo(np.int64).max),
+    ('min_empty_float32', (0, ), np.finfo(np.float32).max),
+    ('min_empty_float64', (0, ), np.finfo(np.float64).max),
     ('sum', (5, 2.0), 10.0),
-    ('sum_initial', (5, 2.0, 2.0), 12.0),
     ('prod', (5, 3.0), 243.0),
-    ('prod_initial', (5, 3.0, 2), 486.0),
 ]
 
 
@@ -48,16 +62,18 @@ def define(omnisci):
         a.fill(v)
         return a.max()
 
-    @omnisci('int8(int32)')
-    def ndarray_max_empty(size):
-        a = Array(size, 'int8')
-        return a.max()
-
-    @omnisci('double(int64, double, double)')
-    def ndarray_max_initial(size, v, initial):
-        a = Array(size, 'double')
-        a.fill(v)
-        return a.max(initial=initial)
+    for retty in NUMERIC_TYPES:
+        for op in ('min', 'max', 'mean'):
+            fn_name = f'ndarray_{op}_empty_{retty}'
+            fn = (f'def {fn_name}(size):\n'
+                  f'    a = Array(size, "{retty}")\n'
+                  f'    return a.{op}()\n')
+            exec(fn)
+            fn = locals()[fn_name]
+            if op == 'mean':
+                omnisci('float64(int32)')(fn)
+            else:
+                omnisci(f'{retty}(int32)')(fn)
 
     @omnisci('double(int64, double)')
     def ndarray_mean(size, v):
@@ -65,74 +81,28 @@ def define(omnisci):
         a.fill(v)
         return a.mean()
 
-    @omnisci('float64(int64)')
-    def ndarray_mean_empty_float(size):
-        a = Array(size, 'float64')
-        return a.mean()
-
-    @omnisci('float64(int64)')
-    def ndarray_mean_empty_int(size):
-        a = Array(size, 'int32')
-        return a.mean()
-
     @omnisci('double(int64, double)')
     def ndarray_min(size, v):
         a = Array(size, 'double')
         a.fill(v)
         return a.min()
 
-    @omnisci('int16(int64)')
-    def ndarray_min_empty(size):
-        a = Array(size, 'int16')
-        return a.min()
-
-    @omnisci('double(int64, double, double)')
-    def ndarray_min_initial(size, v, initial):
-        a = Array(size, 'double')
-        a.fill(v)
-        return a.min(initial=initial)
-
     @omnisci('double(int64, double)')
     def ndarray_sum(size, v):
         a = Array(size, 'double')
         a.fill(v)
         return a.sum()
 
-    @omnisci('double(int64, double, double)')
-    def ndarray_sum_initial(size, v, initial):
-        a = Array(size, 'double')
-        a.fill(v)
-        return a.sum(initial=initial)
-
     @omnisci('double(int64, double)')
     def ndarray_prod(size, v):
         a = Array(size, 'double')
         a.fill(v)
         return a.prod()
 
-    @omnisci('double(int64, double, double)')
-    def ndarray_prod_initial(size, v, initial):
-        a = Array(size, 'double')
-        a.fill(v)
-        return a.prod(initial=initial)
-
 
 @pytest.mark.parametrize("method, args, expected", ndarray_methods,
                          ids=[item[0] for item in ndarray_methods])
 def test_ndarray_methods(omnisci, method, args, expected):
-    if omnisci.has_cuda and omnisci.version < (5, 5):
-        pytest.skip(
-            f'{method}: crashes CUDA enabled omniscidb server [issue 93]')
-
-    if available_version[:3] == (5, 3, 1) and method in ['fill']:
-        pytest.skip(
-            f'{method}: crashes CPU-only omniscidb server v 5.3.1 [issue 113]')
-
-    if available_version[:3] >= (5, 3, 1) and method in ['max_empty']:
-        pytest.skip(
-            f'{method}: fails on CPU-only omniscidb server'
-            ' v 5.3.1+ [issue 114]')
-
     query_args = ', '.join(map(str, args))
     query = f'SELECT ndarray_{method}({query_args})'
 
